#include "tinysvg.h"
#include "xmlutil.h"
#include "artgl.h"

#include <assert.h>
#include <string.h>
#include <ctype.h>
#include <stdio.h>

#include <expat/expat.h>

#include <GL/gl.h>
#include "mzmalloc.h"

static void tinysvg_transform_identity(double t[6]) {
  art_affine_identity(t);
}

static const double* tinysvg_transform_to_4x4(double t[6]) {

  static double m[16] = { 
    0, 0, 0, 0, 
    0, 0, 0, 0, 
    0, 0, 0, 0, 
    0, 0, 0, 0, 
  };

  m[0] = t[0]; m[4] = t[2]; m[12] = t[4];
  m[1] = t[1]; m[5] = t[3]; m[13] = t[5];

  m[10] = 1;
  m[15] = 1;

  return m;
  
}

typedef struct TINYSVGXML TinySvgXml;

struct TINYSVGXML {
  TinySvg* svg;
  TinySvgGroup* curgroup;
};


void tinysvg_tag_svg(TinySvgXml* xmlinfo, const char *el, const char **attr) {

  assert(!xmlinfo->svg);
  assert(!xmlinfo->curgroup);

  TinySvg* rval = mznew(TinySvg, 1);

  xmlinfo->svg = rval;
  xmlinfo->curgroup = tinysvg_group_create(rval, NULL);

}

void tinysvg_tag_group(TinySvgXml* xmlinfo, const char *el, const char **attr) {

  assert(xmlinfo->svg);
  assert(xmlinfo->curgroup);

  xmlinfo->curgroup = tinysvg_group_add(xmlinfo->svg, 
                                        xmlinfo->curgroup, 
                                        xmlutil_lookup_attr(attr, "id"));
  
}

typedef struct TINYSVGPATHD TinySvgPathd;

struct TINYSVGPATHD {
  const char* str;
  ArtBpath* bpath;
  unsigned int capacity;
  unsigned int count;
  double startx;
  double starty;
  double lastx;
  double lasty;
  double coordx;
  double coordy;
};

double tinysvg_pathd_coord(TinySvgPathd* pdinfo) {

  // parse whitespace and/or comma
  while (isspace(pdinfo->str[0])) { ++(pdinfo->str); }
  if (pdinfo->str[0] == ',') { ++(pdinfo->str); }
  while (isspace(pdinfo->str[0])) { ++(pdinfo->str); }

  const char* orig = pdinfo->str;

  if (pdinfo->str[0] == '-') { ++(pdinfo->str); }
  
  if (pdinfo->str[0] == '.') { 
    ++(pdinfo->str);
    if (!isdigit(pdinfo->str[0])) {
      fprintf(stderr, "bad coord!\n");
      exit(-1);
    }
    while (isdigit(pdinfo->str[0])) { ++pdinfo->str; }
  } else if (isdigit(pdinfo->str[0])) {
    // before decimal
    while (isdigit(pdinfo->str[0])) { ++pdinfo->str; }
    if (pdinfo->str[0] == '.') { 
      ++pdinfo->str; 
      while (isdigit(pdinfo->str[0])) { ++pdinfo->str; }
    }
  } else {
    fprintf(stderr, "bad coord!\n");
    exit(-1);
  }
  
  unsigned int len = pdinfo->str - orig;
  char* buf = (char*)alloca(len);
  memcpy(buf, orig, len);
  buf[len] = 0;

  return strtod(buf, NULL);

}

void tinysvg_pathd_coordpair(TinySvgPathd* pdinfo, int rel) {

  double x, y;

  x = tinysvg_pathd_coord(pdinfo);

  y = tinysvg_pathd_coord(pdinfo);

  if (rel) {
    pdinfo->coordx = pdinfo->lastx + x;
    pdinfo->coordy = pdinfo->lasty + y;
    //printf("  relative (%lf, %lf) to (%lf, %lf)\n", x, y, pdinfo->coordx, pdinfo->coordy);
  } else {
    pdinfo->coordx = x;
    pdinfo->coordy = y;
    //printf("  to (%lf, %lf)\n" , pdinfo->coordx, pdinfo->coordy);
  }
  
}

// make sure pdinfo can hold one more than it has
ArtBpath* tinysvg_pathd_expand(TinySvgPathd* pdinfo) {
  assert(pdinfo->count <= pdinfo->capacity);
  if (!pdinfo->capacity) {
    assert(!pdinfo->count);
    assert(!pdinfo->bpath);
    pdinfo->capacity = 8;
    pdinfo->bpath = (ArtBpath*)mzmalloc(pdinfo->capacity*sizeof(ArtBpath));
  } else if (pdinfo->count == pdinfo->capacity) {
    assert(pdinfo->bpath);
    pdinfo->capacity *= 2;
    pdinfo->bpath = (ArtBpath*)mzrealloc(pdinfo->bpath, pdinfo->capacity*sizeof(ArtBpath));
  }
  ++(pdinfo->count);
  return pdinfo->bpath + (pdinfo->count-1);
}

void tinysvg_pathd_record(TinySvgPathd* pdinfo, ArtBpath* p) {
  pdinfo->lastx = p->x3;
  pdinfo->lasty = p->y3;
  if (p->code == ART_MOVETO) {
    pdinfo->startx = p->x3;
    pdinfo->starty = p->y3;
  }
}


void tinysvg_pathd_m(TinySvgPathd* pdinfo, int rel) {
  //printf("moveto\n");
  ArtBpath* p = tinysvg_pathd_expand(pdinfo);
  p->code = ART_MOVETO;
  tinysvg_pathd_coordpair(pdinfo, rel);
  p->x3 = pdinfo->coordx;
  p->y3 = pdinfo->coordy;
  tinysvg_pathd_record(pdinfo, p);
}

void tinysvg_pathd_z(TinySvgPathd* pdinfo) {
  //printf("closepath\n");
  ArtBpath* p = tinysvg_pathd_expand(pdinfo);
  *p = *(pdinfo->bpath);
  p->code = ART_LINETO;
  p->x3 = pdinfo->startx;
  p->y3 = pdinfo->starty;
  tinysvg_pathd_record(pdinfo, p);
}

void tinysvg_pathd_h(TinySvgPathd* pdinfo, int rel) {
  //printf("hline\n");
  ArtBpath* p = tinysvg_pathd_expand(pdinfo);
  p->code = ART_LINETO;
  double val = tinysvg_pathd_coord(pdinfo);
  pdinfo->coordy = pdinfo->lasty;
  if (rel) {
    pdinfo->coordx = pdinfo->lastx + val;
    //printf("  relative (%lf, %lf) to (%lf, %lf)\n", val, 0.0, pdinfo->coordx, pdinfo->coordy);
  } else {
    pdinfo->coordx = val;
    //printf("  to (%lf, %lf)\n", pdinfo->coordx, pdinfo->coordy);
  }
  p->x3 = pdinfo->coordx;
  p->y3 = pdinfo->coordy;
  tinysvg_pathd_record(pdinfo, p);
}

void tinysvg_pathd_v(TinySvgPathd* pdinfo, int rel) {
  //printf("vline\n");
  ArtBpath* p = tinysvg_pathd_expand(pdinfo);
  p->code = ART_LINETO;
  double val = tinysvg_pathd_coord(pdinfo);
  pdinfo->coordx = pdinfo->lastx;
  if (rel) {
    pdinfo->coordy = pdinfo->lasty + val;
    //printf("  relative (%lf, %lf) to (%lf, %lf)\n", 0.0, val, pdinfo->coordx, pdinfo->coordy);
  } else {
    pdinfo->coordy = val;
    //printf("  to (%lf, %lf)\n", pdinfo->coordx, pdinfo->coordy);
  }
  p->x3 = pdinfo->coordx;
  p->y3 = pdinfo->coordy;
  tinysvg_pathd_record(pdinfo, p);
}

void tinysvg_pathd_l(TinySvgPathd* pdinfo, int rel) {
  //printf("lineto\n");
  ArtBpath* p = tinysvg_pathd_expand(pdinfo);
  p->code = ART_LINETO;
  tinysvg_pathd_coordpair(pdinfo, rel);
  p->x3 = pdinfo->coordx;
  p->y3 = pdinfo->coordy;
  tinysvg_pathd_record(pdinfo, p);
}

void tinysvg_pathd_c(TinySvgPathd* pdinfo, int rel) {
  //printf("curveto\n");
  ArtBpath* p = tinysvg_pathd_expand(pdinfo);
  p->code = ART_CURVETO;
  tinysvg_pathd_coordpair(pdinfo, rel);
  p->x1 = pdinfo->coordx;
  p->y1 = pdinfo->coordy;
  tinysvg_pathd_coordpair(pdinfo, rel);
  p->x2 = pdinfo->coordx;
  p->y2 = pdinfo->coordy;
  tinysvg_pathd_coordpair(pdinfo, rel);
  p->x3 = pdinfo->coordx;
  p->y3 = pdinfo->coordy;
  tinysvg_pathd_record(pdinfo, p);
}

void tinysvg_pathd_s(TinySvgPathd* pdinfo, int rel) {

  //printf("shorthand curveto\n");

  ArtBpath* p = tinysvg_pathd_expand(pdinfo);
  const ArtBpath* prev = (p - 1);

  p->code = ART_CURVETO;

  if (prev->code == ART_CURVETO) {
    p->x1 = prev->x3 + (prev->x3 - prev->x2);
    p->y1 = prev->y3 + (prev->y3 - prev->y2);
    //printf("  mirrored prev to (%lf, %lf)\n", p->x1, p->y1);
  } else {
    p->x1 = pdinfo->lastx;
    p->y1 = pdinfo->lasty;
    //printf("  kept prev (%lf, %lf)\n", p->x1, p->y1);
  }

  tinysvg_pathd_coordpair(pdinfo, rel);
  p->x2 = pdinfo->coordx;
  p->y2 = pdinfo->coordy;

  tinysvg_pathd_coordpair(pdinfo, rel);
  p->x3 = pdinfo->coordx;
  p->y3 = pdinfo->coordy;

  tinysvg_pathd_record(pdinfo, p);
  
}

ArtVpath* tinysvg_pathd(const char* str) {

  char c;
  TinySvgPathd working;

  while (isspace(str[0])) { ++str; }

  if (!str || str[0] != 'M') { return 0; }

  memset(&working, 0, sizeof(TinySvgPathd));
  working.str = str;

  while ( (c = working.str[0])) { 
    ++working.str;
    if (isspace(c)) { continue; }
    switch (c) {
    case 'M':
      tinysvg_pathd_m(&working, 0);
      break;
    case 'm':
      tinysvg_pathd_m(&working, 1);
      break;
    case 'H':
      tinysvg_pathd_h(&working, 0);
      break;
    case 'h':
      tinysvg_pathd_h(&working, 1);
      break;
    case 'V':
      tinysvg_pathd_v(&working, 0);
      break;
    case 'v':
      tinysvg_pathd_v(&working, 1);
      break;
    case 'Z': 
    case 'z':
      tinysvg_pathd_z(&working);
      break;
    case 'L':
      tinysvg_pathd_l(&working, 0);
      break;
    case 'l':
      tinysvg_pathd_l(&working, 1);
      break;
    case 'C':
      tinysvg_pathd_c(&working, 0);
      break;
    case 'c':
      tinysvg_pathd_c(&working, 1);
      break;
    case 'S':
      tinysvg_pathd_s(&working, 0);
      break;
    case 's':
      tinysvg_pathd_s(&working, 1);
      break;
    default:
      fprintf(stderr, "unexpected path code %c in path d\n", c);
      exit(-1);
    }
  }

  assert(working.bpath);
  ArtBpath* p = tinysvg_pathd_expand(&working);
  p->code = ART_END;

  //printf("diff is %d.\n", p - working.bpath);

  /*
  int i;
  //printf("bpath:\n");
  for (i=0; working.bpath[i].code != ART_END; ++i) {
    switch(working.bpath[i].code) {
    case ART_MOVETO:
    case ART_MOVETO_OPEN:
      //printf("  moveto (%lf, %lf)\n", working.bpath[i].x1, working.bpath[i].y1);
      break;
    case ART_LINETO:
      //printf("  lineto (%lf, %lf)\n", working.bpath[i].x1, working.bpath[i].y1);
      break;
    case ART_CURVETO:
      //printf("  curveto (%lf, %lf)-(%lf, %lf)-(%lf, %lf)\n", 
             working.bpath[i].x1, working.bpath[i].y1,
             working.bpath[i].x2, working.bpath[i].y2,
             working.bpath[i].x3, working.bpath[i].y3);
      break;
    default:
      break;
    }
  }
  */

  ArtVpath* vpath = art_bez_path_to_vec(working.bpath, 0.25);

  /*
  //printf("vpath:\n");
  for (i=0; vpath[i].code != ART_END; ++i) {
    switch(vpath[i].code) {
    case ART_MOVETO:
    case ART_MOVETO_OPEN:
      //printf("  moveto (%lf, %lf)\n", vpath[i].x, vpath[i].y);
      break;
    case ART_LINETO:
      //printf("  lineto (%lf, %lf)\n", vpath[i].x, vpath[i].y);
      break;
    default:
      break;
    }
  }
  */

  mzfree(working.bpath); 

  return vpath;
  
};



void tinysvg_pathgetcolor(const char* str, const char* attr, unsigned int* value, unsigned int defval) {
  *value = defval;
  if (!str || !attr) { return; }
  char* tmp = strstr(str, attr);
  int tv;
  if (tmp) {
    if (!strncmp(tmp+strlen(attr), "none;", 5)) {
      *value = 0;
    }
    if (sscanf(tmp + strlen(attr), "#%6x", &tv)) {
      *value = ((tv << 8) | 0xff);
      //printf("path %s#%X\n", attr, tv);
    }
  }
}

void tinysvg_pathgetdouble(const char* str, const char* attr, double* value, double defval) {
  *value = defval;
  if (!str || !attr) { return; }
  char* tmp = strstr(str, attr);
  double dv;
  if (tmp) {
    if (sscanf(tmp + strlen(attr), "%lf", &dv)) {
      *value = dv;
      //printf("path %s%lf\n", attr, dv);
    }
  }
}

void tinysvg_pathstyle(TinySvgPath* p, const char* style) {

  tinysvg_pathgetcolor(style, "fill:", &(p->fill_rgba), 0xff); 
  tinysvg_pathgetcolor(style, "stroke:", &(p->stroke_rgba), 0x00); 
  tinysvg_pathgetdouble(style, "stroke-width:", &(p->stroke_width), 1.0); 
  
}

void tinysvg_tag_path(TinySvgXml* xmlinfo, const char *el, const char **attr) {

  ArtVpath* vpath = tinysvg_pathd(xmlutil_lookup_attr(attr, "d"));

  if (!vpath) { return; }

  TinySvgPath* p = tinysvg_path_add(xmlinfo->svg, 
                                    xmlinfo->curgroup, 
                                    xmlutil_lookup_attr(attr, "id"),
                                    vpath);
  
  tinysvg_pathstyle(p, xmlutil_lookup_attr(attr, "style"));
  
}

void tinysvg_tag_start(void *data, const char *el, const char **attr) {

  TinySvgXml* xmlinfo = (TinySvgXml*)data;
  
  if (!strcmp(el, "svg")) {
    tinysvg_tag_svg(xmlinfo, el, attr);
  } else if (!strcmp(el, "g")) {
    tinysvg_tag_group(xmlinfo, el, attr);
  } else if (!strcmp(el, "path")) {
    tinysvg_tag_path(xmlinfo, el, attr);
  } else {
    assert(! "unexpected tag in tinysvg_tag_start");
  }

} 

void tinysvg_tag_end(void *data, const char *el) {
    TinySvgXml* xmlinfo = (TinySvgXml*)data;
    if (!strcmp(el, "g")) {
      assert(xmlinfo->curgroup);
      assert(xmlinfo->curgroup->parent);
      xmlinfo->curgroup = xmlinfo->curgroup->parent;
    }
}  

TinySvg* tinysvg_load(const char* filename) {

  FILE* fp = fopen(filename, "r");
  assert(fp);

  char buf[1024];
  unsigned int rval;
  TinySvgXml xmlinfo;
  xmlinfo.svg = 0;
  xmlinfo.curgroup = 0;

  XML_Parser p = XML_ParserCreate(NULL);
  XML_SetElementHandler(p, tinysvg_tag_start, tinysvg_tag_end);
  XML_SetUserData(p, &xmlinfo);

  while ( (rval = fread(buf, 1, 1024, fp)) ) {
    int pv = XML_Parse(p, buf, rval, rval < 1024);
    if (!pv) {
      fprintf(stderr, "error: %s\n", XML_ErrorString(XML_GetErrorCode(p)));
      exit(-1);
    }
  }
  
  fclose(fp);

  assert(xmlinfo.svg);

  TinySvg* s = xmlinfo.svg;
  s->w = (s->maxx - s->minx);
  s->h = (s->maxy - s->miny);
  s->tx = -(s->minx);
  s->ty = -(s->maxy);
  strncpy(s->filename, strip_filename(filename), 64);

  return s;

}

TinySvgGroup* tinysvg_group_create(TinySvg* t, const char* id) {
  assert(t);
  assert(!t->n_groups);
  TinySvgGroup* rval = mznew(TinySvgGroup, 1);
  tinysvg_transform_identity(rval->transform);
  if (id) { strncpy(rval->id, id, 32); }
  ++(t->n_groups);
  t->root = rval;
  return rval;
}

TinySvgGroup* tinysvg_group_add(TinySvg* t, TinySvgGroup* g, const char* id) {
  assert(t && g);
  TinySvgGroup* rval = mznew(TinySvgGroup, 1);
  tinysvg_transform_identity(rval->transform);
  if (id) { strncpy(rval->id, id, 32); }
  rval->type = TINYSVG_TYPE_GROUP;
  rval->parent = g;
  ++(t->n_groups);
  mzrenew(g->children, void*, g->n_children, g->n_children+1);
  g->children[g->n_children-1] = rval;
  return rval;
}
 
TinySvgPath* tinysvg_path_add(TinySvg* t, TinySvgGroup* g, const char* id, ArtVpath* vp) {
  assert(t && g && vp);
  unsigned int i;
  TinySvgPath* rval = (TinySvgPath*)malloc(sizeof(TinySvgPath));
  assert(rval);
  memset(rval, 0, sizeof(TinySvgPath));
  if (id) { strncpy(rval->id, id, 32); }
  rval->type = TINYSVG_TYPE_PATH;
  rval->parent = g;
  ++(t->n_paths);
  mzrenew(g->children, void*, g->n_children, g->n_children+1);
  g->children[g->n_children-1] = rval;
  for (i=0; vp[i].code!=ART_END; ++i) { 
    if (!t->n_verts || vp[i].x < t->minx) { t->minx = vp[i].x; }
    if (!t->n_verts || vp[i].x > t->maxx) { t->maxx = vp[i].x; }
    if (!t->n_verts || vp[i].y < t->miny) { t->miny = vp[i].y; }
    if (!t->n_verts || vp[i].y > t->maxy) { t->maxy = vp[i].y; }
    ++(t->n_verts);
    ++(rval->n_verts);
  }
  rval->vpath = vp;
  return rval;
}

void tinysvg_free(TinySvg* t) {
  assert(t);
  tinysvg_group_free(t, t->root);
  assert(t->n_groups == 0);
  assert(t->n_paths == 0);
  assert(t->n_verts == 0);
  mzfree(t);
}

void tinysvg_group_free(TinySvg* t, TinySvgGroup* g) {
  assert(t && g);
  unsigned int i;
  for (i=0; i<g->n_children; ++i) {
    TinySvgNode* n = (TinySvgNode*)(g->children[i]);
    switch (n->type) {
    case TINYSVG_TYPE_GROUP:
      tinysvg_group_free(t, (TinySvgGroup*)n);
      break;
    case TINYSVG_TYPE_PATH:
      tinysvg_path_free(t, (TinySvgPath*)n);
      break;
    default:
      fprintf(stderr, "Bad type in SVG node!\n");
      exit(-1);
    }
  }
  --(t->n_groups);
  if (g->n_children) {
    assert(g->children);
    mzfree(g->children);
  }
  mzfree(g);
}

void tinysvg_path_free(TinySvg* t, TinySvgPath* p) {
  assert(t && p);
  art_free(p->vpath);
  --(t->n_paths);
  t->n_verts -= p->n_verts;
  mzfree(p);
}

static unsigned int tinysvg_color_invert(unsigned int rgba) {
  unsigned char r = (rgba >> 24) & 0xff;
  unsigned char g = (rgba >> 16) & 0xff;
  unsigned char b = (rgba >>  8) & 0xff;
  unsigned char a = (rgba      ) & 0xff;
  r = 255 - r;
  g = 255 - g;
  b = 255 - b;
  return (r << 24) | (g << 16) | (b << 8) | a;
}

static void tinysvg_path_draw(TinySvgPath* p, int invert) {
  assert(p && p->vpath);
  if (p->fill_rgba & 0xff) {
    unsigned int color = p->fill_rgba;
    if (invert) { color = tinysvg_color_invert(color); }
    artgl_vpath_alpha(p->vpath, color);
  }
  if (p->stroke_rgba & 0xff && p->stroke_width) {
    unsigned int color = p->stroke_rgba;
    if (invert) { color = tinysvg_color_invert(color); }
    artgl_vpath_stroke_alpha(p->vpath,
                             color,
                             ART_PATH_STROKE_JOIN_ROUND,
                             ART_PATH_STROKE_CAP_ROUND,
                             p->stroke_width);
  }
}

static void tinysvg_group_draw(TinySvgGroup* g, int invert) {
  assert(g);
  unsigned int i;
  glPushMatrix();
  glMultMatrixd(tinysvg_transform_to_4x4(g->transform));
  for (i=0; i<g->n_children; ++i) {
    TinySvgNode* n = (TinySvgNode*)(g->children[i]);
    switch (n->type) {
    case TINYSVG_TYPE_GROUP:
      tinysvg_group_draw((TinySvgGroup*)n, invert);
      break;
    case TINYSVG_TYPE_PATH:
      tinysvg_path_draw((TinySvgPath*)n, invert);
      break;
    default:
      fprintf(stderr, "Bad type in SVG node!\n");
      exit(-1);
    }
  }
  glPopMatrix();
}

void tinysvg_draw(TinySvg* t, int flip, int invert) {
  assert(t && t->root);
  glPushMatrix();
  glTranslated(t->tx, t->ty, 0);
  if (flip) {
    glTranslated(0.5*t->w, 0, 0);
    glScaled(-1, 1, 1);
    glTranslated(-0.5*t->w, 0, 0);
  }
  tinysvg_group_draw(t->root, invert);
  glPopMatrix();
}
